import mock
import unittest2 as unittest

from eh import constants
from eh import exceptions as exc
from eh.tests import base_test as base
from eh import topic as top


class TestTopic(base.TestCase):
    def setUp(self):
        super(TestTopic, self).setUp()

        self.good_data = """[//]: # (testing) This is test contents
# Something about TESTING

Lorem ipsum something dolar emet. Yup.
"""
        self.conf = {}
        self.root_node = {}
        self.rootpath = '/home/place/'
        self.path = "parent/topic.md"
        self.meta = None
        open_name = 'eh.topic.open'
        self.patch1 = mock.patch(
            open_name, mock.mock_open(
                read_data=self.good_data))
        self.open_mock = self.patch1.start()
        self.topic = top.Topic(
            self.conf, self.root_node, self.rootpath, self.path)
        self.good_graph = {
            constants.PARENT_KEY: {
                "parent": { constants.TOPIC_KEY: ["poop"] }
            },
            "parent": { constants.TOPIC_KEY: ["poop"] }
        }
        self.good_graph_no_parents = {
            constants.PARENT_KEY: { },
            constants.TOPIC_KEY: ["poop"]
        }

    def tearDown(self):
        super(TestTopic, self).tearDown()
        mock.patch.stopall()

    def test_create_topic(self):
        self.assertIsNotNone(self.topic)

    def test_create_topic_removed_preamble(self):
        text_without_top_line = "\n".join(self.good_data.splitlines()[1:])
        self.assertEqual(len(self.topic.text), len(text_without_top_line))

    def test_create_topic_meta_correct(self):
        self.assertEqual(self.topic.meta[0], 'testing')

    def test_create_topic_subject_correct(self):
        self.assertEqual(self.topic.summary, 'This is test contents')

    def test_create_topic_key_exists(self):
        self.assertIsNotNone(self.topic.key)

    def test_create_topic_invalid_root_raises(self):
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, None, self.rootpath, self.path)
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, "poop", self.rootpath, self.path)
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, 2, self.rootpath, self.path)
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, 2.0, self.rootpath, self.path)
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, 0, self.rootpath, self.path)
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, ['asdf'], self.rootpath, self.path)
        with self.assertRaises(exc.TopicStoreInvalidRoot):
            topic = top.Topic(
                self.conf, ('asdf',), self.rootpath, self.path)

    def test_create_topic_invalid_rootpath_raises(self):
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, None, self.path)
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, 1.0, self.path)
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, 1, self.path)
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, 0, self.path)
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, ['asdf'], self.path)
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, {'asdf':'asdf'}, self.path)
        with self.assertRaises(exc.TopicInvalidRootPath):
            topic = top.Topic(
                self.conf, {}, ('asdf',), self.path)

    def test_is_topic(self):
        self.assertTrue(self.topic.is_topic('parent/topic'))
        self.assertFalse(self.topic.is_topic('topic'))
        self.assertFalse(self.topic.is_topic('parent topic'))
        self.assertFalse(self.topic.is_topic(None))

    def test_shortkey(self):
        self.assertTrue(self.topic.shortkey == 'topic')

    def test_remove_comment_preamble(self):
        text_top_line = self.good_data.splitlines()[0]
        self.assertEquals(
            top.Topic.remove_comment_preamble(text_top_line),
            text_top_line[len(constants.COMMENT_PREAMBLE)+1:])

    def test_split_meta_from_summary(self):
        text_top_line = self.good_data.splitlines()[0]
        cleaned = top.Topic.remove_comment_preamble(text_top_line)
        meta, summary = top.Topic.split_meta_from_summary(cleaned)

        self.assertEquals(['testing'], meta)
        self.assertEquals('This is test contents', summary)

    def test_check_comment_format(self):
        self.assertTrue(top.Topic.check_comment_format(self.good_data))
        self.assertFalse(top.Topic.check_comment_format(
            constants.COMMENT_PREAMBLE + " missing meta tags"))
        self.assertFalse(top.Topic.check_comment_format(
            constants.COMMENT_PREAMBLE + " (missing, summary)"))
        self.assertFalse(top.Topic.check_comment_format(
            constants.COMMENT_PREAMBLE + " () empty meta"))
        self.assertFalse(top.Topic.check_comment_format(None))

    def test_parse_topic_contents(self):
        with mock.patch(
                'eh.topic.open',
                mock.mock_open(read_data=self.good_data)) as mock_open:
            meta, summary, text = top.Topic.parse_topic_contents(
                self.conf, self.root_node, self.rootpath, self.path)
            self.assertEquals(['testing'], meta)
            self.assertEquals('This is test contents', summary)

            text_without_top_line = "\n".join(self.good_data.splitlines()[1:])
            self.assertEquals(text_without_top_line, text)

    @unittest.skip("TODO: learn to mock missing files")
    def test_parse_topic_contents_not_found(self):
        with mock.patch(
                'eh.topic.open',
                mock.mock_open(read_data=None)) as mock_open:
            meta, summary, text = top.Topic.parse_topic_contents(
                self.conf, self.root_node, self.rootpath, self.path)
            self.assertIsNone(meta)
            self.assertIsNone(summary)
            self.assertIsNone(text)

    def test_parse_topic_contents_invalid_preamble_raises(self):
        bad = """[/]: # (testing) This is test contents
# Something about TESTING

Lorem ipsum something dolar emet. Yup.
"""
        with mock.patch(
                'eh.topic.open',
                mock.mock_open(read_data=bad)) as mock_open:
            with self.assertRaises(exc.TopicError):
                top.Topic.parse_topic_contents(
                    self.conf, self.root_node, self.rootpath, self.path)

    def test_parse_topic_contents_missing_preamble_raises(self):
        bad = """(testing) This is test contents
# Something about TESTING

Lorem ipsum something dolar emet. Yup.
"""
        with mock.patch(
                'eh.topic.open',
                mock.mock_open(read_data=bad)) as mock_open:
            with self.assertRaises(exc.TopicError):
                top.Topic.parse_topic_contents(
                    self.conf, self.root_node, self.rootpath, self.path)

    def test_parse_topic_contents_missing_meta_raises(self):
        bad = """[//]: # This is test contents
# Something about TESTING

Lorem ipsum something dolar emet. Yup.
"""
        with mock.patch(
                'eh.topic.open',
                mock.mock_open(read_data=bad)) as mock_open:
            with self.assertRaises(exc.TopicError):
                top.Topic.parse_topic_contents(
                    self.conf, self.root_node, self.rootpath, self.path)

    def test_map_root(self):
        root = {}
        top.Topic.map_topic_path_to_root_node({}, root, self.path, 'poop')
        self.assertEquals(self.good_graph, root)

    def test_repr(self):
        self.assertEquals(
            "parent/topic 65 chars ['testing'] This is test contents",
            str(self.topic))

    def test_map_root(self):
        root = {}
        path = 'topic.md'
        top.Topic.map_topic_path_to_root_node({}, root, path, 'poop')
        self.assertEquals(self.good_graph_no_parents, root)



import os

from eh import constants
from eh import exceptions as exc
from eh import topic_key as tk

class Topic(object):
    def __init__(self, conf, root_node, rootpath, path):
        self.path = path
        if root_node is None or not isinstance(root_node, dict):
            raise exc.TopicStoreInvalidRoot()
        self.root_node = root_node
        if rootpath is None or not isinstance(rootpath, str) or not rootpath:
            raise exc.TopicInvalidRootPath()
        self.rootpath = rootpath
        self.meta, self.summary, self.text = (
            Topic.parse_topic_contents(conf, root_node, rootpath, path))
        self.key = tk.TopicKey(conf, self.path, self.meta)
        Topic.map_topic_path_to_root_node(conf, root_node, path, self)

    def is_topic(self, topic):
        if self.key.matches(topic):
            return True
        return False
    
    @property
    def shortkey(self):
        return self.key.shortkey

    def __repr__(self):
        return constants.STR_TOPIC_REPR % (
            self.key, len(self.text), self.meta, self.summary)

    @staticmethod
    def map_topic_path_to_root_node(conf, root_node, path, node):
        path_parts = path.split(constants.KEY_DIVIDE_CHAR)
        current_node = root_node
        if constants.PARENT_KEY not in root_node:
            root_node[constants.PARENT_KEY] = {}
        for i in range(len(path_parts)):
            if i == len(path_parts) - 1:  # leaf nodes
                if constants.TOPIC_KEY not in current_node:
                    current_node[constants.TOPIC_KEY] = []
                current_node[constants.TOPIC_KEY].append(node)
            else:
                p = path_parts[i]
                if p not in current_node:
                    current_node[p] = {}
                current_node = current_node[p]
                parent_path = constants.KEY_DIVIDE_CHAR.join(path_parts[0:i+1])
                if parent_path not in root_node[constants.PARENT_KEY]:
                    root_node[constants.PARENT_KEY][parent_path] = current_node

    @staticmethod
    def parse_topic_contents(conf, root_node, root, path):
        fullpath = os.path.join(root, path)
        with open(fullpath, 'r') as myfile:
            data = myfile.read().splitlines()
            top_line = data.pop(0)
            if not Topic.check_comment_format(top_line):
                raise exc.TopicError("Missing expected preamble")
            text = constants.CR_CHAR.join(data)
            meta, summary = Topic.split_meta_from_summary(top_line)
            return meta, summary, text
        return None, None, None  # pragma: no cover

    @staticmethod
    def remove_comment_preamble(text):
        """
        Removes the comment preamble from a single line of text and returns
        the rest.
        """
        return text.replace(
            constants.COMMENT_PREAMBLE, constants.EMPTY).strip()

    @staticmethod
    def split_meta_from_summary(text):
        """
        Returns the meta tags and summary as tuple from comment line

        Input such as:
            (tag1, tag2) Summary...

        Would return:
            (["tag1", "tag2"]), "Summary..."
        """
        text = Topic.remove_comment_preamble(text)
        meta = text[0:text.index(constants.META_END_CHAR)+1]
        text = text.replace(meta, constants.EMPTY).strip()
        meta = meta.replace(constants.META_END_CHAR, constants.EMPTY).replace(
            constants.META_START_CHAR, constants.EMPTY)
        meta_list = [m.strip() for m in meta.split(constants.META_DIVIDE_CHAR)]
        return meta_list, text

    @staticmethod
    def check_comment_format(text):
        """
        Checks a single line to see if it satisfies requirements to be a real
        comment
        """
        if not text or not isinstance(text, str):
            return False
        if not text or not text.startswith(constants.COMMENT_PREAMBLE):
            return False
        text = Topic.remove_comment_preamble(text)
        if (
                not text.startswith(constants.META_START_CHAR)
                or constants.META_END_CHAR not in text):
            return False
        meta, text = Topic.split_meta_from_summary(text)
        return any(meta) and (text)



class EhError(Exception):
    pass


class TopicError(EhError):
    pass


class TopicStoreInvalidRoot(EhError):
    pass


class TopicInvalidRootPath(EhError):
    pass


class InvalidValue(EhError):
    pass


class UnimplementedFeature(EhError):
    pass



import mock

from eh import constants
from eh import exceptions as exc
from eh.tests import base_test as base
from eh import topic_store as ts


class TopicMock(mock.Mock):
    def __init__(self, *args, **kwargs):
        super(TopicMock, self).__init__(*args, **kwargs)
        self.topic = 'topic'

    def is_topic(self, topic_key):
        return topic_key == self.topic


class TestTopicStore(base.TestCase):
    def setUp(self):
        super(TestTopicStore, self).setUp()
        self.conf = {
            "stuff": True
        }
        self.topics = {
            "topic1": {},
            "parent": {
                "topic2": {},
                "topic3": {},
            }
        }
        self.topic2 = TopicMock()
        self.topic2.topic = "parent/topic2"
        self.topic1 = TopicMock()
        self.topic1.topic = "topic1"
        self.root = {
            constants.PARENT_KEY: ["parent"],
            constants.TOPIC_KEY: [self.topic1],
            "parent": { constants.TOPIC_KEY: [self.topic2] }
        }

    def tearDown(self):
        super(TestTopicStore, self).tearDown()
        mock.patch.stopall()

    def test_create_store(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())

    def test_invalid_topic_format_continues(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        with mock.patch('eh.topic.Topic.parse_topic_contents') as mock_parse:
            mock_parse.side_effect = exc.TopicError
            store._parse_topics({}, {}, "/", ['a', 'b'])

    def test_get_topic_valid_path(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        store.root_node = self.root
        topic = store.get_topic('parent/topic2')
        self.assertIsNotNone(topic)

    def test_get_topic_simple_path(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        store.root_node = self.root
        topic = store.get_topic('topic1')
        self.assertIsNotNone(topic)

    def test_get_topic_invalid_path(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        store.root_node = self.root
        topic = store.get_topic('topic2')
        self.assertIsNone(topic)

    def test_default_store_does_not_update(self):
        store = ts.TopicStore({}, "")
        ret = store.update()
        self.assertIsNone(ret)

    def test_gather_topics_empty_dir(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        with mock.patch('os.listdir', return_value=[]) as list_mock:
            topics = store._gather_topics(self.conf, '/', '/')
            self.assertEquals(0, len(topics))
            self.assertEquals(0, store.topic_count())
            self.assertIsNotNone(topics)

    def test_gather_topics_one_item_no_ext(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        with mock.patch('os.listdir', return_value=["thing"]) as list_mock:
            topics = store._gather_topics(self.conf, '/', '/')
            self.assertEquals(0, len(topics))
            self.assertIsNotNone(topics)

    def test_gather_topics_one_item_with_ext(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        with mock.patch('os.listdir', return_value=["thing.md"]) as list_mock:
            topics = store._gather_topics(self.conf, '/', '/')
            self.assertEquals(1, len(topics))
            self.assertIsNotNone(topics)

    def test_gather_topics_removes_mainpath(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        with mock.patch('os.listdir', return_value=["thing.md"]) as list_mock:
            topics = store._gather_topics(self.conf, '/a/b/c', '/a/b/c/d')
            self.assertEquals(1, len(topics))
            self.assertIsNotNone(topics)
            self.assertTrue('/a/b/c' not in topics[0])

    def test_gather_topics_mixed_items(self):
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())
        mixed_list = ["poop", "thing.md"]
        with mock.patch('os.listdir', return_value=mixed_list) as list_mock:
            topics = store._gather_topics(self.conf, '/', '/')
            self.assertEquals(1, len(topics))
            self.assertIsNotNone(topics)

    @mock.patch('os.path.isdir')
    @mock.patch('os.listdir')
    def test_gather_topics_with_directory(
            self, mock_listdir, mock_isdir):
        mixed_list = ["a_dir", "thing.md"]
        mock_isdir.side_effect = lambda x : x == '/a_dir'
        mock_listdir.side_effect = (
            lambda x : mixed_list if x != 'a_dir' else 'thing2.md')
        store = ts.TopicStore({}, "")
        self.assertEquals(0, store.topic_count())

        topics = store._gather_topics(self.conf, '/', '/')
        self.assertEquals(2, len(topics))
        self.assertIsNotNone(topics)

    @mock.patch('eh.topic.Topic', new_callable=TopicMock)
    def test_parse_topics_returned_topics_equals_paths(
            self, mock_topic):
        store = ts.TopicStore(self.conf, "")
        self.assertEquals(0, store.topic_count())
        root = {}
        main_path = '/'
        paths = [
            'path1',
            'path2'
        ]
        topics = store._parse_topics(self.conf, root, main_path, paths)
        self.assertEquals(len(paths), len(topics))
        self.assertEquals(len(paths), mock_topic.call_count)

    @mock.patch('eh.topic.Topic', new_callable=TopicMock)
    def test_parse_topics_empty_paths_okay(
            self, mock_topic):
        store = ts.TopicStore(self.conf, "")
        self.assertEquals(0, store.topic_count())
        root = {}
        main_path = '/'
        paths = []
        topics = store._parse_topics(self.conf, root, main_path, paths)
        self.assertEquals(len(paths), len(topics))
        self.assertEquals(len(paths), mock_topic.call_count)

    @mock.patch('eh.topic.Topic.parse_topic_contents')
    def test_parse_topics_invalid_root_fails(self, mock_ptc):
        mock_ptc.return_value = (None, None, None)
        store = ts.TopicStore(self.conf, "")
        self.assertEquals(0, store.topic_count())
        main_path = '/'
        paths = [
            'path1',
            'path2'
        ]
        for _t in [None, "", []]:
            with self.assertRaises(exc.TopicStoreInvalidRoot):
                topics = store._parse_topics(self.conf, _t, main_path, paths)

    def test_parse_topics_invalid_path_fails(self):
        store = ts.TopicStore(self.conf, "")
        self.assertEquals(0, store.topic_count())
        root = {}
        paths = [
            'path1',
            'path2'
        ]
        for _t in [None, "", [], {}]:
            with self.assertRaises(exc.TopicInvalidRootPath):
                topics = store._parse_topics(self.conf, root, _t, paths)

    def test_get_filtered_parent_keys(self):
        parent_node = {
            constants.PARENT_KEY: "something",
            constants.TOPIC_KEY: "something else",
            "Some key": "another something",
        }
        filtered = ts.TopicStore.get_filtered_parent_keys(parent_node)
        self.assertEquals(1, len(filtered))
        self.assertFalse(constants.PARENT_KEY in filtered)
        self.assertFalse(constants.TOPIC_KEY in filtered)
        self.assertTrue("Some key" in filtered)

    def test_get_filtered_parent_keys_empty_dict_okay(self):
        parent_node = {}
        filtered = ts.TopicStore.get_filtered_parent_keys(parent_node)
        self.assertEquals(0, len(filtered))

    def test_get_filtered_parent_keys_weird_values(self):
        for _t in [None, [], "asdf"]:
            with self.assertRaises(exc.InvalidValue):
                ts.TopicStore.get_filtered_parent_keys(_t)

    def test_get_topics_for_parent_returns_root_with_none(self):
        store = ts.TopicStore(self.conf, "")
        store.root_node = self.root
        self.assertEquals(0, store.topic_count())
        with mock.patch('eh.topic_store.TopicStore._parents') as mock_parents:
            mock_parents.return_value = {}
            root_ret = store.get_topics_for_parent(None)
            self.assertEquals(([self.topic1], ['parent']), root_ret)

    def test_get_topics_for_parent_returns_none_when_not_found(self):
        store = ts.TopicStore(self.conf, "")
        self.assertEquals(0, store.topic_count())
        with mock.patch('eh.topic_store.TopicStore._parents') as mock_parents:
            mock_parents.return_value = {}
            ret = store.get_topics_for_parent("woo")
            self.assertEquals((None, None), ret)

    def test_has_parent(self):
        store = ts.TopicStore(self.conf, "")
        store.root_node = self.root
        self.assertTrue(store.has_parent("parent"))
        self.assertFalse(store.has_parent("topic1"))
        self.assertFalse(store.has_parent("nothing"))

    def test_parents(self):
        store = ts.TopicStore(self.conf, "")
        store.root_node = self.root
        self.assertIsNotNone(store._parents())
        self.assertEquals(1, len(store._parents()))

    @mock.patch('eh.topic.Topic.is_topic')
    def test_has_topic(self, mock_is_topic):
        store = ts.TopicStore(self.conf, "")
        t1 = TopicMock()
        t1.topic = 'poo'

        store._topics = [t1]
        self.assertTrue(store.has_topic('poo'))
        self.assertFalse(store.has_topic('foo'))



from eh import constants

class TopicKey(object):
    def __init__(self, conf, path, meta):
        self.path = path
        self.meta = meta
        TopicKey.parse_topic_path(conf, self.path)
        self.keyparts = TopicKey.parse_topic_path(conf, self.path)

    def matches(self, topic_string):
        if topic_string == str(self):
            return True
        return False

    @property
    def shortkey(self):
        return self.keyparts[-1]

    @staticmethod
    def parse_topic_path(conf, path):
        key = path
        for ext in constants.KNOWN_EXT:
            if key.endswith(ext):
                key = "".join(key.rsplit(ext, 1))
        keyparts = key.split(constants.KEY_DIVIDE_CHAR)
        return keyparts

    def __repr__(self):
        return constants.KEY_DIVIDE_CHAR.join(self.keyparts)




import mdv

from prettytable import PrettyTable
from prettytable import PLAIN_COLUMNS, NONE

from eh import constants


class Output(object):
    def __init__(self, conf):
        self.conf = conf

    def _topic_table(self, topics):
        t = PrettyTable(
            ['Subject', 'Summary'],
            padding_width=0,
            style=PLAIN_COLUMNS,
            vertical_char=' ', horizontal_char=' ', junction_char=' ',
            hrules=NONE)
        t.align['Subject'] = 'l'
        t.align['Summary'] = 'l'
        for topic in topics:
            t.add_row([topic.key, topic.summary])
        return t

    def _parent_table(self, top_level, parents, manager):
        t = PrettyTable(
            ['Subtopics', 'Topics', 'Samples'],
            padding_width=0,
            style=PLAIN_COLUMNS,
            vertical_char=' ', horizontal_char=' ', junction_char=' ',
            hrules=NONE)
        t.align['Subtopics'] = 'l'
        t.align['Topics'] = 'l'
        t.align['Samples'] = 'l'
        for parent in parents:
            full_parent = parent
            if top_level:
                full_parent = constants.KEY_DIVIDE_CHAR.join([
                    top_level, parent])
            tops, subtops = manager.get_topics_for_parent(full_parent)
            tops_list = []
            for top in tops:
                tops_list.append("%s" % top.shortkey)
            top_str = ", ".join(tops_list)
            t.add_row(
                [full_parent, "%d" % len(tops), top_str])
        return t

    def output_list(self, top_level, topics, parents, manager):
        top_level_str = "" if not top_level else (
            "For topic %s " % top_level)
        findings = "I have found %d topics and %d subtopics:" % (
                len(topics), len(parents))
        table_str = str(self._topic_table(topics))

        out = "%s%s\n%s" % (top_level_str, findings, table_str)
        if parents:
            out += "\n" +  " " * 30
            out += "\n%s" % self._parent_table(top_level, parents, manager)
        return out


class MarkdownOutput(Output):
    def __init__(self, conf):
        super(MarkdownOutput, self).__init__(conf)
        self.conf = conf
        self.no_colors = False

    def output_topic(self, topic):
        pre_md = topic.text
        md = mdv.main(pre_md, no_colors=self.no_colors)
        lines = md.splitlines()
        lines = [line for line in lines if line.strip()]
        return "\n".join(lines)



import os

from eh import constants
from eh import exceptions as exc
from eh import topic

class TopicStore(object):
    """
    The TopicStore class contains the root node of the topic graph.

    All topics can be mapped from the root node. The code to produce the graph
    is in the Topic module.

    The graph structure is used to store the parent/child relationships in
    topics. It allows for easy lookup and display of options.

    The root node is a dict. It looks like this:

        root = {
            constants.PARENT_KEY: ["parent"],
            constants.TOPIC_KEY: ["topic1"],
            "parent": { constants.TOPIC_KEY: ["topic2"] }
        }

    - constants.PARENT_KEY provides a list of all parents within the graph
    - constants.TOPIC_KEY provides all topics on the same level (in this case
      root)
    - every key afterwards is the parent and its value is a dict with the same
      structure relative to the same level (in this case root) except they will
      not have the PARENT_KEY

    self._topics is a flattened list of all topic objects in this store

    """
    def __init__(self, conf, filepath):
        self.root_node = {}
        self.root_node[constants.PARENT_KEY] = {}
        self.filepath = filepath
        self.topic_paths = []
        self._topics = []

    def initialize(self, conf):
        self.topic_paths = self._gather_topics(
            conf, self.filepath, self.filepath)
        self._topics = self._parse_topics(
            conf, self.root_node, self.filepath, self.topic_paths)

    def update(self):
        pass

    def _gather_topics(self, conf, mainpath, filepath):
        """
        Given paths locate files with known extensions and return as list.

        This method is recursive and will search subdirectories.

        On initial call mainpath and filepath will likely be the same. The 
        mainpath will be the root of the topic store, and the filepath will
        hold the current subdirectory being searched.

        All files will have their paths returned relative to mainpath (this
        means that mainpath shouldn't be in the return).
        """
        found_topics = []
        if not conf or not mainpath or not filepath:
            return found_topics
        for f in os.listdir(filepath):
            full_path = os.path.join(filepath, f)
            if (
                    not os.path.isdir(full_path) and 
                    f.lower().endswith(tuple(constants.KNOWN_EXT))):
                found_topics.append(full_path)
            if os.path.isdir(full_path):
                found_topics.extend(self._gather_topics(
                    conf, mainpath, full_path))
        found_topics = [
            p.replace(mainpath, constants.EMPTY) for p in found_topics]
        return found_topics

    def _parse_topics(self, conf, root_node, rootpath, topic_paths):
        """
        Produce all topic objects given a list of paths and store in graph
        structure whose root is represented by root_node.

        rootpath: is a common path shared by everything in the store and is
        not used for topic 'keying' (the absolute path leading up to where
        the topic store begins)

        topic_paths: is a list of topic_paths starting from rootpath. this path
        is used for topic 'keying'

        returns: a list of all topic objects created during this process
        """
        topics = []
        for p in topic_paths:
            try:
                t = topic.Topic(conf, root_node, rootpath, p)
                topics.append(t)
            except exc.TopicError:
                continue
        return topics

    def _parents(self):
        """
        Return a list of all parents for this store.

        All parents (subdirectories) encountered are stored in a list on the
        root node.
        """
        return self.root_node[constants.PARENT_KEY]

    def topic_count(self):
        return len(self._topics)

    def has_topic(self, topic):
        for t in self._topics:
            if t.is_topic(topic):
                return True
        return False

    def has_parent(self, parent):
        return parent in self._parents()

    def get_topic(self, path):
        """
        Return the topic object given a path.

        Locate the specific topic object by traversing the graph using the
        path. It walks the path until it reaches the 'level' where the topic
        node should be.
        """
        path_parts = path.split(constants.KEY_DIVIDE_CHAR)
        node = self.root_node
        if len(path_parts) > 1:
            for k in path_parts[:-1]:
                node = node[k]

        for topic in node[constants.TOPIC_KEY]:
            if topic.is_topic(path):
                return topic
        return None
    
    def get_topics_for_parent(self, parent):
        """
        Given a parent, the root if parent is None, return topics and subtopics
        """
        p = self.root_node
        if parent:
            p = self._parents().get(parent, None)
        if p is None:
            return None, None
        return p.get(
            constants.TOPIC_KEY, []), TopicStore.get_filtered_parent_keys(p)

    @staticmethod
    def get_filtered_parent_keys(parent_node):
        """
        Produces a list of keys (sub-directories and topics) for a parent.
        """
        drop_list = [constants.PARENT_KEY, constants.TOPIC_KEY]
        try:
            return [k for k in parent_node.keys() if k not in drop_list]
        except AttributeError:
            raise exc.InvalidValue(
                "Provided parent node is invalid (expected a dict)")



import os
import shutil

from ConfigObject import ConfigObject

from eh import constants


def is_true(value):
    return value.upper() in ['TRUE', 'T', 'YES']


def open_config():
    if not os.path.exists(constants.CONF_DIR):
        os.mkdir(constants.CONF_DIR)
    if not os.path.exists(constants.CONF_FILE):
        shutil.copyfile(
            constants.DEFAULT_CONF,
            constants.CONF_FILE)
    return ConfigObject(filename=constants.CONF_FILE)



from ConfigObject import ConfigObject
import mock

from eh.tests import base_test as base
from eh import topic_manager as tm
from eh import topic_store as ts


class StoreMock(mock.Mock):
    def __init__(self, *args, **kwargs):
        super(StoreMock, self).__init__(*args, **kwargs)
        self.topic_name = 'topic'

    def has_topic(self, topic):
        if topic == 'pass':
            return True
        return False

    def get_topic(self, topic):
        if topic == 'pass':
            return self.topic_name
        return None

    def has_parent(self, parent):
        if parent == 'pass':
            return True
        return False

    def get_topics_for_parent(self, parent):
        if parent == 'pass' or parent == "":
            return [self.topic_name], ['parent']
        return [], []


class TestTopicManager(base.TestCase):
    def setUp(self):
        super(TestTopicManager, self).setUp()
        self.test_conf = ConfigObject()
        self.test_conf['topic_stores'] = {
            'thing': 'some/directory'
        }
        self.patch1 = mock.patch(
            'eh.topic_store.TopicStore', new_callable=StoreMock)
        self.store = self.patch1.start()
        self.store.topic_name = 'topic1'
        self.patch2 = mock.patch(
            'eh.topic_store.TopicStore', new_callable=StoreMock)
        self.store2 = self.patch2.start()
        self.store2.topic_name = 'topic2'

    def tearDown(self):
        mock.patch.stopall()

    def test_initialize_empty_conf_okay(self):
        manager = tm.TopicManager({})
        self.assertEqual(0, len(manager.topic_stores))

    def test_initialize_none_conf_okay(self):
        manager = tm.TopicManager(None)
        self.assertEqual(0, len(manager.topic_stores))

    def test_initialize_with_test_okay(self):
        manager = tm.TopicManager(self.test_conf)
        self.assertEqual(1, len(manager.topic_stores))

    def test_has_topic_matches_internal_store(self):
        manager = tm.TopicManager(self.test_conf)
        self.assertTrue(manager.has_topic('pass'))
        self.assertFalse(manager.has_topic('fail'))

    def test_has_topic_no_stores_has_no_match(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = []
        self.assertFalse(manager.has_topic('pass'))
        manager.topic_stores = [self.store]
        self.assertTrue(manager.has_topic('pass'))

    def test_get_topic_matches_internal_store(self):
        manager = tm.TopicManager(self.test_conf)
        self.assertIsNotNone(manager.get_topic('pass'))
        self.assertIsNone(manager.get_topic('fail'))

    def test_get_topic_no_stores_has_no_match(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = []
        self.assertIsNone(manager.get_topic('pass'))
        manager.topic_stores = [self.store]
        self.assertIsNotNone(manager.get_topic('pass'))

    def test_has_parent_matches_internal_store(self):
        manager = tm.TopicManager(self.test_conf)
        self.assertTrue(manager.has_parent('pass'))
        self.assertFalse(manager.has_parent('fail'))

    def test_has_parent_no_stores_has_no_match(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = []
        self.assertFalse(manager.has_parent('pass'))
        manager.topic_stores = [self.store]
        self.assertTrue(manager.has_parent('pass'))

    def test_get_topics_for_parent_matches_internal_store(self):
        manager = tm.TopicManager(self.test_conf)
        ret, pret = manager.get_topics_for_parent('pass')
        self.assertEqual(1, len(ret))
        self.assertEqual(1, len(pret))

    def test_get_topics_for_parent_no_store(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = []
        ret, pret = manager.get_topics_for_parent('pass')
        self.assertEqual(0, len(ret))
        self.assertEqual(0, len(pret))

        manager.topic_stores = [self.store]
        ret, pret = manager.get_topics_for_parent('pass')
        self.assertEqual(1, len(ret))
        self.assertEqual(1, len(pret))

    def test_get_topics_for_parent_multiple_stores(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = [self.store, self.store2]
        ret, pret = manager.get_topics_for_parent('pass')
        self.assertEqual(2, len(ret))
        self.assertEqual(2, len(pret))

    def test_get_root_list(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = [self.store, self.store2]
        ret, pret = manager.get_root_list()
        self.assertEqual(2, len(ret))
        self.assertEqual(2, len(pret))

    def test_get_root_list_empty_with_no_stores(self):
        manager = tm.TopicManager(self.test_conf)
        manager.topic_stores = []
        ret, pret = manager.get_root_list()
        self.assertEqual(0, len(ret))
        self.assertEqual(0, len(pret))



import os

from eh import constants
from eh import topic_store as ts
from eh import git_store as gs


class TopicManager(object):
    def __init__(self, conf):
        self.topic_stores = self._gather_topics(conf)

    def _topic_store_path(self, name):
        return os.path.join(constants.CONF_DIR, name)

    def _gather_topics(self, conf):
        out = []
        if not conf:
            return out
        for name, location in conf.items(constants.CONF_TOPIC_STORE):
            if location.startswith('https://github.com/'):
                t_store = gs.GitTopicStore(
                    conf, location, self._topic_store_path(name))
            else:
                t_store = ts.TopicStore(conf, location)
            t_store.initialize(conf)
            out.append(t_store)
        return out

    def update(self):
        for s in self.topic_stores:
            s.update()

    def has_topic(self, topic):
        for s in self.topic_stores:
            if s.has_topic(topic):
                return True
        return False

    def get_topic(self, topic):
        for s in self.topic_stores:
            if s.has_topic(topic):
                return s.get_topic(topic)
        return None

    def has_parent(self, parent):
        for s in self.topic_stores:
            if s.has_parent(parent):
                return True
        return False

    def get_topics_for_parent(self, parent):
        ret = []
        pret = []
        for s in self.topic_stores:
            topics, parents = s.get_topics_for_parent(parent)
            if topics:
                ret.extend(topics)
            if parents:
                pret.extend(parents)
        return ret, pret

    def get_root_list(self):
        return self.get_topics_for_parent("")



import os

COMMENT_PREAMBLE = '[//]: #'
META_START_CHAR = '('
META_END_CHAR = ')'
META_DIVIDE_CHAR = ','
CR_CHAR = '\n'
KNOWN_EXT = ['.md']
KEY_DIVIDE_CHAR = os.sep
EMPTY = ''
CONF_TOPIC_STORE = 'topic_stores'
CONF_DIR_NAME = '.eh'
CONF_NAME = 'eh.ini'
TOPIC_KEY = "_"
PARENT_KEY = "_parents"
STR_TOPIC_REPR = "%s %d chars %s %s" 

ROOT_DIR = os.path.dirname(os.path.abspath(__file__)) 
DEFAULT_STORE = os.path.join(ROOT_DIR, 'contrib/subjects/')
PACKAGE_DIR = os.path.join(ROOT_DIR, "..")
USERHOME = os.path.expanduser('~')
CONF_DIR = os.path.join(USERHOME, CONF_DIR_NAME)
CONF_FILE = os.path.join(CONF_DIR, CONF_NAME)
DEFAULT_CONF = os.path.join(PACKAGE_DIR, 'default_conf.ini')



import mock

from eh import constants
from eh import exceptions as exc
from eh.tests import base_test as base
from eh import topic_key as tk


class TestTopicKey(base.TestCase):
    def setUp(self):
        super(TestTopicKey, self).setUp()
        self.conf = {}
        self.path = "parent/topic.md"
        self.meta = None

    def tearDown(self):
        super(TestTopicKey, self).tearDown()
        mock.patch.stopall()

    def test_create_key(self):
        key = tk.TopicKey(self.conf, self.path, self.meta)
        self.assertIsNotNone(key)

    def test_matches(self):
        key = tk.TopicKey(self.conf, self.path, self.meta)
        self.assertTrue(key.matches('parent/topic'))
        self.assertFalse(key.matches('topic'))
        self.assertFalse(key.matches('parent'))
        self.assertFalse(key.matches('topic/parent'))

    def test_short_key(self):
        key = tk.TopicKey(self.conf, self.path, self.meta)
        self.assertEqual('topic', key.shortkey)
        self.assertNotEqual('parent', key.shortkey)
        self.assertNotEqual('parent/topic', key.shortkey)

    def test_parse_topic_path(self):
        keyparts = tk.TopicKey.parse_topic_path(self.conf, "simple")
        self.assertIsNotNone(keyparts)
        self.assertEquals(1, len(keyparts))
        self.assertEquals("simple", keyparts[0])

        keyparts = tk.TopicKey.parse_topic_path(self.conf, "parent/simple")
        self.assertIsNotNone(keyparts)
        self.assertEquals(2, len(keyparts))
        self.assertEquals("parent", keyparts[0])
        self.assertEquals("simple", keyparts[1])

        with mock.patch('eh.constants.KNOWN_EXT', ['.ext']) as mock_ext:
            keyparts = tk.TopicKey.parse_topic_path(self.conf, "simple.ext")
            self.assertIsNotNone(keyparts)
            self.assertEquals(1, len(keyparts))
            self.assertEquals("simple", keyparts[0])

    def test_parse_topic_paths_ignore_internal_extension(self):
        with mock.patch('eh.constants.KNOWN_EXT', ['.ext']) as mock_ext:
            keyparts = tk.TopicKey.parse_topic_path(self.conf, "whoa.ext.ext")
            self.assertIsNotNone(keyparts)
            self.assertEquals(1, len(keyparts))
            self.assertEquals("whoa.ext", keyparts[0])

    def test_topic_key_repr(self):
        key = tk.TopicKey(self.conf, self.path, self.meta)
        self.assertTrue(key.matches('parent/topic'))
        self.assertTrue('parent/topic', str(key))




import sys
from eh.cli import Eh


def completion_hook(cmd, curr_word, prev_word):
    eh_obj = Eh(False, True)
    eh_keys = list(eh_obj.subjects.keys()) + list(eh_obj.repo_subs.subjects)
    eh_keys.sort()
    matches = [k for k in eh_keys if k.startswith(curr_word)]
    return matches


def main():
    results = completion_hook(*sys.argv[1:])
    if len(results):
          print("\n".join(results))


if __name__ == "__main__":
    main()



# Copyright (c) 2016 Justin L. Hammond
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import click 
import ConfigObject as conf

from eh import config
from eh import constants
from eh import git_store as gs
from eh import output
from eh import topic_manager as tm


command_settings = {
    'ignore_unknown_options': True,
}


@click.command(context_settings=command_settings)
@click.argument('subject', nargs=-1)
@click.option('--debug', is_flag=True)
@click.option('--no-colors', is_flag=True)
@click.pass_context
def main(context, subject, debug, no_colors):
    """
    Eh is a terminal program that will provide you with
    quick reminders about a subject.

    To get started run: eh help

    To figure out what eh knows about run: eh list

    To update the list of subjects: eh update

    Note:

    Eh will make a directory in your userhome called .eh
    where it will store downloaded subjects.
    """

    conf = config.open_config()
    if config.is_true(conf.eh.show_default):
        conf[constants.CONF_TOPIC_STORE][
            'eh_default'] = constants.DEFAULT_STORE


    topic_key = constants.KEY_DIVIDE_CHAR.join(subject)
    manager = tm.TopicManager(conf)
    out = output.MarkdownOutput(conf)

    test_git = gs.GitTopicStore(
            conf, "https://github.com/roaet/eh_subjects", 
            "/home/jhammond/.eh/eh_subjects")

    if topic_key == "list" or len(subject) == 0:
        topics, parents = manager.get_root_list()
        click.echo(out.output_list("", topics, parents, manager))
    elif topic_key == "update":
        manager.update()
    else:
        if manager.has_topic(topic_key):
            topic = manager.get_topic(topic_key)
            click.echo(out.output_topic(topic))
        elif manager.has_parent(topic_key):
            topics, parents = manager.get_topics_for_parent(topic_key)
            click.echo(out.output_list(
                topic_key, topics, parents, manager))



import os
import git

import click

from eh import topic_store as ts


class GitTopicStore(ts.TopicStore):
    def __init__(self, conf, repo, filepath):
        """
        repo - is the git url
        filepath - is the place the repo will be cloned
        """
        self.repo = repo
        self.repo_path = filepath
        super(GitTopicStore, self).__init__(conf, filepath)

    def initialize(self, conf):
        if not self._check_if_directories_exist():
            self._init_repo()
        super(GitTopicStore, self).initialize(conf)

    def _setup_repo_directory(self):
        if not os.path.exists(self.repo_path):
            os.makedirs(self.repo_path)
        git.Repo.clone_from(self.repo, self.repo_path)

    def _init_repo(self):
        try:
            self._setup_repo_directory()
        except git.exc.GitCommandError as e:
            click.echo("Failed to clone: %s" % e)
            exit(1)

    def _get_subjects_from_repo(self):
        if not self._check_if_directories_exist():
            click.echo('Need to initialize subjects')
            self._init_repo()

    def _check_if_directories_exist(self):
        return os.path.exists(self.repo_path)

    def update(self):
        if not self._check_if_directories_exist():
            click.echo('Need to initialize subjects')
            self._init_repo()
        else:
            g = git.cmd.Git(self.repo_path)
            g.pull()
            click.echo("Updated subjects")



import git
import mock

from eh import constants
from eh import exceptions as exc
from eh.tests import base_test as base
from eh import git_store as gs


class TestGitStore(base.TestCase):
    def setUp(self):
        super(TestGitStore, self).setUp()
        self.store = gs.GitTopicStore({}, "somerepo", "")

    def tearDown(self):
        super(TestGitStore, self).tearDown()
        mock.patch.stopall()

    def test_create_git_store(self):
        self.assertIsNotNone(self.store)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('os.makedirs', create=True)
    @mock.patch('git.Repo.clone_from', create=True)
    def test_setup_repo_dir_make_if_missing(
            self, mock_clone, mock_dirs, mock_exists):
        mock_exists.return_value = False
        self.store._setup_repo_directory()
        self.assertEquals(1, mock_dirs.call_count)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('os.makedirs', create=True)
    @mock.patch('git.Repo.clone_from', create=True)
    def test_setup_repo_dir_dont_make_if_there(
            self, mock_clone, mock_dirs, mock_exists):
        mock_exists.return_value = True
        self.store._setup_repo_directory()
        self.assertEquals(0, mock_dirs.call_count)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('os.makedirs', create=True)
    @mock.patch('git.Repo.clone_from', create=True)
    @mock.patch('eh.git_store.exit', create=True)
    def test_init_repo_exit_on_giterror(
            self, mock_exit, mock_clone, mock_dirs, mock_exists):
        mock_exists.return_value = True
        mock_clone.side_effect = git.exc.GitCommandError("clone", "failboat")
        self.store._init_repo()
        self.assertEquals(1, mock_exit.call_count)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('os.makedirs', create=True)
    @mock.patch('git.Repo.clone_from', create=True)
    @mock.patch('eh.git_store.exit', create=True)
    def test_init_repo_no_exit_on_okay(
            self, mock_exit, mock_clone, mock_dirs, mock_exists):
        mock_exists.return_value = True
        self.store._init_repo()
        self.assertEquals(0, mock_exit.call_count)

    @mock.patch('os.path.exists', create=True)
    def test_check_if_dir_exists_based_on_file(self, mock_exists):
        mock_exists.return_value = False
        self.assertFalse(self.store._check_if_directories_exist())
        mock_exists.return_value = True
        self.assertTrue(self.store._check_if_directories_exist())

    @mock.patch('os.path.exists', create=True)
    @mock.patch('eh.git_store.GitTopicStore._init_repo', create=True)
    def test_get_subjects_init_repo_if_missing(self, mock_init, mock_exists):
        mock_exists.return_value = False
        self.store._get_subjects_from_repo()
        self.assertEquals(1, mock_init.call_count)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('eh.git_store.GitTopicStore._init_repo', create=True)
    def test_get_subjects_dont_init_repo_if_there(
            self, mock_init, mock_exists):
        mock_exists.return_value = True
        self.store._get_subjects_from_repo()
        self.assertEquals(0, mock_init.call_count)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('eh.git_store.GitTopicStore._init_repo', create=True)
    @mock.patch('git.cmd.Git', create=True)
    def test_update_pull_if_exists(self, mock_git, mock_init, mock_exists):
        mock_exists.return_value = True
        mock_pull = mock.MagicMock()
        mock_git_cmd = mock.MagicMock()
        mock_git_cmd.pull = mock_pull
        mock_git.return_value = mock_git_cmd
        self.store.update()
        self.assertEquals(1, mock_git.call_count)
        self.assertEquals(1, mock_pull.call_count)
        self.assertEquals(0, mock_init.call_count)

    @mock.patch('os.path.exists', create=True)
    @mock.patch('eh.git_store.GitTopicStore._init_repo', create=True)
    @mock.patch('git.cmd.Git', create=True)
    def test_update_init_if_new(self, mock_git, mock_init, mock_exists):
        mock_exists.return_value = False
        mock_pull = mock.MagicMock()
        mock_git_cmd = mock.MagicMock()
        mock_git_cmd.pull = mock_pull
        mock_git.return_value = mock_git_cmd
        self.store.update()
        self.assertEquals(0, mock_git.call_count)
        self.assertEquals(0, mock_pull.call_count)
        self.assertEquals(1, mock_init.call_count)
